//go:build !fasthttp
// +build !fasthttp

package routebuildermdl

import (
	"context"
	"io/ioutil"
	"net"
	"net/http"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/authmdl/jwtmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/authmdl/roleenforcemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/servicebuildermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/statemdl"

	"github.com/gin-gonic/gin"
	version "github.com/hashicorp/go-version"
	"github.com/tidwall/gjson"
)

// Init routing init
func Init(o, r, c *gin.RouterGroup, JWTKey string) {
	o.POST("/mql/state", statemdl.StateHandler)
	o.POST("/mql", OpenHandler)
	r.POST("/mql", RestrictedHandler)
	c.POST("/mql", RoleBasedHandler)
	o.POST("/heavymql", HeavyOpenHandler)
	r.POST("/heavymql", HeavyRestrictedHandler)
	c.POST("/heavymql", HeavyRoleBasedHandler)
	o.GET("/checkAPIStatus", func(c *gin.Context) {
		c.String(http.StatusOK, "OK")
	})
	jwtmdl.GlobalJWTKey = JWTKey
}

func commonHandler(c *gin.Context, isRestricted, isRoleBased, heavyDataActivity bool, principalObj servicebuildermdl.Principal) {
	serviceHeader := c.Request.Header.Get("Service-Header")
	branch := CleanBranch(
		strings.TrimSpace(c.Request.Header.Get(Header_Branch)),
	)
	services := strings.Split(serviceHeader, ",")
	versionError := appVersioning(c)
	if versionError != nil {
		c.JSON(http.StatusExpectationFailed, versionError.Error())
		return
	}
	responseMap := make(map[string]responseData)
	responseDataObj := responseData{}
	var reqBody []byte

	var readError error
	reqBody, readError = ioutil.ReadAll(c.Request.Body)
	if errormdl.CheckErr2(readError) != nil {
		responseDataObj.Error = errormdl.CheckErr2(readError).Error()
		responseDataObj.ErrorCode = errormdl.EXPECTATIONFAILED
		loggermdl.LogError(readError)
		c.JSON(http.StatusExpectationFailed, responseDataObj)
		return
	}

	requestBody := gjson.ParseBytes(reqBody)
	for i := 0; i < len(services); i++ {
		responseDataObj := responseData{}
		service := services[i]
		result, ab, isCompressed, errorCode, err := executeServiceWithBranch(service, branch, []byte(requestBody.Get(service).String()), isRestricted, isRoleBased, heavyDataActivity, principalObj)
		if errormdl.CheckErr1(err) != nil {
			if ab == nil {
				responseDataObj.ErrorCode = errorCode
				responseDataObj.Error = err.Error()
			} else {
				responseDataObj.Error = ab.GetErrorData()
				if responseDataObj.Error == nil {
					responseDataObj.Error = err.Error()
				}
				errorCode := ab.GetErrorCode()
				if errorCode == 0 {
					errorCode = errormdl.EXPECTATIONFAILED
				}
				responseDataObj.ErrorCode = errorCode
				if ab.TransactionEnable {
					loggermdl.LogError("transaction enabled rollback")

					var err error
					// database transaction rollback if transaction is enabled
					switch ab.DatabaseType {
					case dalmdl.MYSQL:
						if ab.TXN != nil {
							loggermdl.LogError("MYSQL Transaction Rollbacked")
							err = ab.TXN.Rollback()
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.MYSQLERROR
							}
						}

					case dalmdl.SQLSERVER:
						if ab.SQLServerTXN != nil {
							loggermdl.LogError("SQLSERVER Transaction Rollbacked")
							err = ab.SQLServerTXN.Rollback()
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.MYSQLERROR
							}
						}

					case dalmdl.GraphDB:
						if ab.GraphDbTXN != nil {
							loggermdl.LogError("GRAPHDB Transaction Rollbacked")
							err = ab.GraphDbTXN.Discard(context.TODO())
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.GRAPHDBERROR
							}
						}

					default:
						loggermdl.LogError("Invalid database type while rollback transaction")

					}

				}
			}
		} else {

			if ab != nil {
				if ab.TransactionEnable {
					var err error

					switch ab.DatabaseType {
					case dalmdl.MYSQL:
						if ab.TXN != nil {
							loggermdl.LogError("MYSQL Transaction Commit")
							err = ab.TXN.Commit()
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.MYSQLERROR
							} else {
								responseDataObj.Result = result
								responseDataObj.ErrorCode = errormdl.NOERROR
							}
						}

					case dalmdl.SQLSERVER:
						if ab.SQLServerTXN != nil {
							loggermdl.LogError("SQLSERVER Transaction Commit")
							err = ab.SQLServerTXN.Commit()
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.SQLSERVERERROR
							} else {
								responseDataObj.Result = result
								responseDataObj.ErrorCode = errormdl.NOERROR
							}
						}

					case dalmdl.GraphDB:
						if ab.SQLServerTXN != nil {
							loggermdl.LogError("GRAPHDB Transaction Commit")
							err = ab.GraphDbTXN.Commit(context.TODO())
							if err != nil {
								responseDataObj.Error = err.Error()
								responseDataObj.ErrorCode = errormdl.GRAPHDBERROR
							} else {
								responseDataObj.Result = result
								responseDataObj.ErrorCode = errormdl.NOERROR
							}
						}

					default:
						loggermdl.LogError("Invalid database type while commit transaction")

					}
				} else {
					responseDataObj.Result = result
					responseDataObj.ErrorCode = errormdl.NOERROR
				}
			} else {
				responseDataObj.Result = result
				responseDataObj.ErrorCode = errormdl.NOERROR
			}
		}
		responseDataObj.IsCompressed = isCompressed
		responseDataObj = formatResponse(ab, responseDataObj)
		responseMap[service] = responseDataObj
		// Token extraction
		if ab != nil {
			token, _ := ab.GetDataString("MQLToken")
			c.Header("Authorization", token)
			if ab.EnableResponseHeaders {
				c.Header("serverIP", ab.GetServerIP())
			}
		}
	}
	c.JSON(http.StatusOK, responseMap)
	return
}

// OpenHandler for /o
func OpenHandler(c *gin.Context) {
	principal := servicebuildermdl.Principal{}
	principal.ClientIP = getClientIP(c)
	commonHandler(c, false, false, false, principal)
}

// RestrictedHandler for /r
func RestrictedHandler(c *gin.Context) {
	pricipalObj, extractError := extractPricipalObject(c)
	if extractError != nil {
		loggermdl.LogError(extractError)
		c.JSON(http.StatusExpectationFailed, extractError.Error())
		return
	}
	commonHandler(c, true, false, false, pricipalObj)
}

// RoleBasedHandler for /r/c
func RoleBasedHandler(c *gin.Context) {
	pricipalObj, extractError := extractPricipalObject(c)
	if extractError != nil {
		loggermdl.LogError(extractError)
		c.JSON(http.StatusExpectationFailed, extractError.Error())
		return
	}
	commonHandler(c, true, true, false, pricipalObj)
}

// HeavyOpenHandler for /o
func HeavyOpenHandler(c *gin.Context) {
	principal := servicebuildermdl.Principal{}
	principal.ClientIP = getClientIP(c)
	commonHandler(c, false, false, true, principal)
}

// HeavyRestrictedHandler for /r
func HeavyRestrictedHandler(c *gin.Context) {
	pricipalObj, extractError := extractPricipalObject(c)
	if extractError != nil {
		loggermdl.LogError(extractError)
		c.JSON(http.StatusExpectationFailed, extractError.Error())
		return
	}
	commonHandler(c, true, false, true, pricipalObj)
}

// HeavyRoleBasedHandler for /r/c
func HeavyRoleBasedHandler(c *gin.Context) {
	pricipalObj, extractError := extractPricipalObject(c)
	if extractError != nil {
		loggermdl.LogError(extractError)
		c.JSON(http.StatusExpectationFailed, extractError.Error())
		return
	}
	commonHandler(c, true, true, true, pricipalObj)
}

func appVersioning(c *gin.Context) error {
	if isAppVersionEnabled {
		appVersion := c.Request.Header.Get("app-version")
		if appVersion == "" {
			return errormdl.Wrap("No App version Found in request header")
		}
		ver, err := version.NewVersion(appVersion)
		if errormdl.CheckErr(err) != nil {
			return errormdl.CheckErr(err)
		}
		if isStrictMode {
			if !ver.Equal(applicationVersion) {
				return errormdl.Wrap("Application version mismatched")
			}
		} else {
			if ver.GreaterThan(applicationVersion) {
				return errormdl.Wrap("Server Version is outdated")
			}
			if ver.LessThan(minimumSupportedVersion) {
				return errormdl.Wrap("Client Version is outdated")
			}
		}
	}
	return nil
}

func extractPricipalObject(c *gin.Context) (servicebuildermdl.Principal, error) {
	principal := servicebuildermdl.Principal{}
	if jwtmdl.GlobalJWTKey == "" {
		return principal, errormdl.Wrap("No Global JWT key found")
	}
	claim, decodeError := jwtmdl.DecodeToken(c.Request)
	if errormdl.CheckErr(decodeError) != nil {
		return principal, errormdl.CheckErr(decodeError)
	}

	groups, grperr := roleenforcemdl.GetGroupNames(claim, "groups")
	if errormdl.CheckErr(grperr) != nil {
		loggermdl.LogError(grperr)
		return principal, errormdl.CheckErr(grperr)
	}
	userID, _ := claim["userId"].(string)
	if len(userID) < 2 {
		loggermdl.LogError("UserID length is less than 2")
		return principal, errormdl.Wrap("UserID length is less than 2")
	}

	rawMetadata, ok := claim["metadata"]
	if ok {
		metadata, ok := rawMetadata.(string)
		if !ok {
			loggermdl.LogError("Unable to parse metadata from JWT Token")
			return principal, errormdl.Wrap("Unable to parse metadata from JWT Token")
		}
		principal.Metadata = metadata
	}
	principal.Groups = groups
	principal.UserID = userID
	principal.Token = c.Request.Header.Get("Authorization")
	principal.ClientIP = getClientIP(c)
	return principal, nil
}

// getClientIP - returns respected header value from request header
func getClientIP(c *gin.Context) string {
	clientIP := c.Request.Header.Get("X-Real-Ip")
	if clientIP == "" {
		clientIP = c.Request.Header.Get("X-Forwarded-For")
	}
	if clientIP == "" {
		clientIP, _, splitHostPortError := net.SplitHostPort(c.Request.RemoteAddr)
		if splitHostPortError == nil && isCorrectIP(clientIP) {
			return clientIP
		}
		return ""
	}
	if isCorrectIP(clientIP) {
		return clientIP
	}
	return ""
}

// isCorrectIP - return true if ip string is valid textual representation of an IP address, else returns false
func isCorrectIP(ip string) bool {
	return net.ParseIP(ip) != nil
}
